//////////////////////////////////////////////////////////////////////////
// spieleprogrammierer.de-Programmier-Contest #09                       //
// ******************************************************************** //
// Titel: ...... "Rechtecke"                                            //
// Typ: ........ Geschwindigkeit                                        //
// Deadline: ... 12.10.2008                                             //
// Abgabe: ..... contest@spieleprogrammierer.de                         //
// ******************************************************************** //
//                                                                      //
// Aufgabenstellung siehe:                                              //
// http://www.spieleprogrammierer.de/phpBB2/viewtopic.php?t=10234       //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <iomanip>
#include "utils.h"

//////////////////////////////////////////////////////////////////////////

// maximale Anzahl von Rechtecken
const uint32_t MAX_NUM_RECTANGLES = 10000000;

// minimale Bildgre
const uint32_t MIN_IMAGE_SIZE = 128;

// maximale Bildgre
const uint32_t MAX_IMAGE_SIZE = 4096;

//////////////////////////////////////////////////////////////////////////

// ein Rechteck
struct Rectangle
{
    Rectangle()
    {
    }

    Rectangle(int32_t begin_x,
              int32_t begin_y,
              int32_t end_x,
              int32_t end_y)
        : begin_x(begin_x)
        , begin_y(begin_y)
        , end_x(end_x)
        , end_y(end_y)
    {
    }

    bool contains(int32_t x,
                  int32_t y) const
    {
        return x >= begin_x && y >= begin_y && x < end_x && y < end_y;
    }

    int32_t begin_x;    // Beginn links (gehrt zum Rechteck)
    int32_t begin_y;    // Beginn oben (gehrt zum Rechteck)
    int32_t end_x;      // Ende rechts (gehrt nicht mehr zum Rechteck, grer als begin_x)
    int32_t end_y;      // Ende unten (gehrt nicht mehr zum Rechteck, grer als begin_y)
};

// Funktionszeigertyp
typedef void (* PainterFunctionPointer)(const Rectangle*, uint32_t, uint32_t*, uint32_t, int32_t, int32_t);

//////////////////////////////////////////////////////////////////////////

// Referenzimplementierung
void reference_painter(const Rectangle* p_rectangles,
                       uint32_t num_rectangles,
                       uint32_t* p_image,
                       uint32_t image_size,
                       int32_t view_x,
                       int32_t view_y)
{
    // alle Rechtecke durchgehen
    for(uint32_t i = 0; i != num_rectangles; ++i)
    {
        // Rechteck holen
        const Rectangle& r = p_rectangles[i];

        // sichtbaren Teil des Rechtecks bestimmen
        const int32_t begin_x = std::max(0, r.begin_x - view_x);
        const int32_t begin_y = std::max(0, r.begin_y - view_y);
        const int32_t end_x = std::min(static_cast<int32_t>(image_size), r.end_x - view_x);
        const int32_t end_y = std::min(static_cast<int32_t>(image_size), r.end_y - view_y);

        // Ist berhaupt etwas sichtbar?
        if(begin_x >= end_x ||
           begin_y >= end_y ||
           begin_x >= static_cast<int32_t>(image_size) ||
           begin_y >= static_cast<int32_t>(image_size) ||
           end_x <= 0 ||
           end_y <= 0)
        {
            // Das Rechteck ist auerhalb des sichtbaren Bereichs.
            continue;
        }

        // alle sichtbaren Pixel des Rechtecks durchgehen und zeichnen
        for(int32_t x = begin_x; x != end_x; ++x)
        {
            for(int32_t y = begin_y; y != end_y; ++y)
            {
                p_image[y * image_size + x] = i;
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////

// Vorbedingungen:
// - p_rectangles ist 16-Byte-ausgerichtet.
// - num_rectangles liegt zwischen 1 und MAX_NUM_RECTANGLES.
// - p_image ist 16-Byte-ausgerichtet.
// - image_size ist eine Zweierpotenz und liegt zwischen MIN_IMAGE_SIZE und MAX_IMAGE_SIZE.
// - Das erste Rechteck im Array berdeckt immer das gesamte Bild ("Hintergrundfarbe" ist 0).
//
// Nachbedingung:
// Jedes Pixel in p_image enthlt den Index des in p_rectangles am weitesten hinten liegenden
// Rechtecks, das dieses Pixel berdeckt.

// >>> DEINE FUNKTION - BITTE IMPLEMENTIEREN! <<<
void my_painter(const Rectangle* p_rectangles,
                uint32_t num_rectangles,
                uint32_t* p_image,
                uint32_t image_size,
                int32_t view_x,
                int32_t view_y)
{
}

//////////////////////////////////////////////////////////////////////////

// generiert zufllige Rechtecke
void generate_rectangles(Rectangle* p_rectangles,
                         uint32_t num_rectangles,
                         uint32_t image_size,
                         MTRand& rng)
{
    assert(num_rectangles && num_rectangles <= MAX_NUM_RECTANGLES);
    assert(p_rectangles);
    assert(image_size >= MIN_IMAGE_SIZE && image_size <= MAX_IMAGE_SIZE);

    // Das erste Rechteck deckt alles ab.
    p_rectangles[0] = Rectangle(0 - image_size, 0 - image_size, image_size, image_size);

    // Die brigen Rechtecke sind zufllig.
    for(uint32_t i = 1; i != num_rectangles; ++i)
    {
        Rectangle& r = p_rectangles[i];

        // Start zufllig whlen
        r.begin_x = 0 - image_size + rng() % (2 * image_size);
        r.begin_y = 0 - image_size + rng() % (2 * image_size);

        // Rechtecktyp bestimmen (nach einer gewissen Verteilung)
        const uint32_t distribution[] = {0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3};
        switch(distribution[rng() % (sizeof(distribution) / 4)])
        {
        case 0:
            // komplett zuflliges Rechteck
            r.end_x = r.begin_x + 1 + rng() % (image_size - r.begin_x);
            r.end_y = r.begin_y + 1 + rng() % (image_size - r.begin_y);
            break;
        case 1:
            // nur ein Punkt
            r.end_x = r.begin_x + 1;
            r.end_y = r.begin_y + 1;
            break;
        case 2:
            // horizontale Linie
            r.end_x = r.begin_x + 1 + rng() % (image_size - r.begin_x);
            r.end_y = r.begin_y + 1;
            break;
        case 3:
            // vertikale Linie
            r.end_x = r.begin_x + 1;
            r.end_y = r.begin_y + 1 + rng() % (image_size - r.begin_y);
            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////

// misst die Geschwindigkeit einer Lsung
double time_function(PainterFunctionPointer function,
                     const Rectangle* p_rectangles,
                     uint32_t num_rectangles,
                     uint32_t* p_image,
                     uint32_t image_size,
                     int32_t view_x,
                     int32_t view_y,
                     double min_test_duration)
{
    const double t0 = time();
    double t = 0.0;
    uint32_t num_tests_done = 0;

    do
    {
        function(p_rectangles, num_rectangles, p_image, image_size, view_x, view_y);
        t = time() - t0;
        ++num_tests_done;
    }
    while(t < min_test_duration);

    return t / num_tests_done;
}

//////////////////////////////////////////////////////////////////////////

// prft, ob die Lsung richtig ist
bool check_result(const Rectangle* p_rectangles,
                  uint32_t num_rectangles,
                  const uint32_t* p_image,
                  uint32_t image_size,
                  int32_t view_x,
                  int32_t view_y)
{
    for(int32_t x = view_x; x != static_cast<int32_t>(view_x + image_size); ++x)
    {
        for(int32_t y = view_y; y != static_cast<int32_t>(view_y + image_size); ++y)
        {
            // Rechteck holen, das angeblich dieses Pixel fllt
            const uint32_t id = p_image[(y - view_y) * image_size + x - view_x];

            // Liegt dieses Rechteck hier?
            if(id >= num_rectangles) return false;
            if(!p_rectangles[id].contains(x, y)) return false;

            // Kein Rechteck mit hherem Index darf hier liegen.
            for(uint32_t i = id + 1; i != num_rectangles; ++i) if(p_rectangles[i].contains(x, y)) return false;
        }
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////

int main()
{
    MTRand rng(4711);

    // Wert erhhen, um genauere Zeitmessungen zu erhalten
    const double MIN_TEST_DURATION = 2.0;

    // Parameter fr die Testlufe
    // (es wird nicht alles getestet, weil das sehr lange dauern wrde)
    const uint32_t TEST_MIN_NUM_RECTANGLES = 1000;
    const uint32_t TEST_MAX_NUM_RECTANGLES = 1000000;
    const uint32_t TEST_MIN_IMAGE_SIZE = 128;
    const uint32_t TEST_MAX_IMAGE_SIZE = 1024;

	std::cout.setf(std::ios::left);
	std::cout << std::setw(12) << "image_size";
	std::cout << std::setw(16) << "num_rectangles";
	std::cout << std::setw(6) << "OK?";
	std::cout << std::setw(14) << "Zeit (ms)";
	std::cout << std::setw(19) << "Referenzzeit (ms)";
	std::cout << std::setw(10) << "Speed-up";
	std::cout << std::endl;
	std::cout << std::string(77, '-').c_str();
	std::cout << std::endl;

    Rectangle* p_rectangles = aligned_alloc<Rectangle>(TEST_MAX_NUM_RECTANGLES);
    uint32_t* p_image = aligned_alloc<uint32_t>(TEST_MAX_IMAGE_SIZE * TEST_MAX_IMAGE_SIZE);

    for(uint32_t image_size = TEST_MIN_IMAGE_SIZE; image_size <= TEST_MAX_IMAGE_SIZE; image_size *= 2)
    {
        for(uint32_t num_rectangles = TEST_MIN_NUM_RECTANGLES; num_rectangles <= TEST_MAX_NUM_RECTANGLES; num_rectangles *= 10)
        {
            std::cout << std::setw(12) << image_size;
            std::cout << std::setw(16) << num_rectangles;

            // Rechtecke generieren
            generate_rectangles(p_rectangles, num_rectangles, image_size, rng);

            // zufllige Ansicht
            const int32_t view_x = 0 - image_size + rng() % image_size;
            const int32_t view_y = 0 - image_size + rng() % image_size;

            // Funktion ausfhren und Zeit messen
            const double t_my = time_function(&my_painter, p_rectangles, num_rectangles, p_image, image_size, view_x, view_y, MIN_TEST_DURATION);

            // auf Korrektheit testen
            const bool correct = check_result(p_rectangles, num_rectangles, p_image, image_size, view_x, view_y);
            std::cout << std::setw(6) << (correct ? "Ja" : "Nein");

            // Zeit ausgeben
            std::cout << std::setw(14) << (1000.0 * t_my);

            // Referenzfunktion ausfhren und Zeit messen
            const double t_reference = time_function(&reference_painter, p_rectangles, num_rectangles, p_image, image_size, view_x, view_y, MIN_TEST_DURATION);
            std::cout << std::setw(19) << (1000.0 * t_reference);

            // Speed-up ermitteln
            std::cout << std::setw(10) << (t_reference / t_my);
            std::cout << std::endl;
        }
    }

    aligned_free(p_rectangles);
    aligned_free(p_image);

    return 0;
}
